/*
 * bitonic.sc - Split-C bitonic sort program
 */
#include <stdio.h>
#include <split-c/split-c.h>
#include "sort.h"
#include "local_radix.h"
#include "transpose.h"
#include "bitonic.h"

/*
 * Statically allocate NP_MAX keys of buffer space on each processor
 */
#define NP_MAX (1024*1024)
int data[PROCS]::[NP_MAX];
int temp[PROCS]::[NP_MAX];

/*
 * Pointers to the local buffer space
 */
int *localData;
int *localTemp;

/* 
 * Time for individual phases
 */
double start, finish;
double sort_time[50], total_sort; 
double trans_time[50], total_trans;
double untrans_time[50], total_untrans;
double merge_time[50], total_merge;
double scatter_time[50], total_scatter;
double gather_time;

int logPROCS;

/*
 * Types to describe the function of a processor when simulating the
 * butterfly, the position of a key in a bitonic sequence, and the
 * shape of a bitonic sequence. 
 */
typedef enum {COMP_MIN, COMP_MAX} comp_t;
typedef enum {UP_POINT, DOWN_POINT, UP_MID, DOWN_MID, UNKNOWN_DIR} dir_point_t;
typedef enum {SINGLE_DOWN, SINGLE_UP, DOUBLE_DOWN, DOUBLE_UP, SLOPE_DOWN, 
                SLOPE_UP, IDENTICAL} bitonic_t;

void splitc_main(int argc, char **argv) 
{
  int c;
  extern char *optarg;
  int np = (1024*16);          
  int distrib = 0, i, procs;

  /* Get options from the user */
  while ((c = getopt(argc, argv, "n:d:h")) != -1) {
    switch(c) {
    case 'h':
      on_one {
        printf("Bitonic sort options: (defaults)\n");
        printf("  -n number of keys (in K) per processor  (64K)\n");
        printf("  -d key distribution\n");
      }
      break;
    case 'n':  np = atoi(optarg);  np *= 1024; break;
    case 'd':  distrib = atoi(optarg); break;
    }
  }
  on_one  {
    printf("Bitonic sort of %d keys on %d processors\n", np, PROCS);
    printf("Key distribution: %d.\n", distrib);
  }

  /* compute log_2 of the number of processors.  Must be an integer. */
  for (i = 1, logPROCS = 0; i < PROCS; i = i << 1, logPROCS++);
  if (i != PROCS) {
    printf("Couldn't calculate log of PROCS.  Exiting.\n");
    exit(1);
  }

  localData = (int *) data;
  localTemp = (int *) temp; 
  for (i = 0; i < 50; i++) {
    trans_time[i] = 0;
    untrans_time[i] = 0;
    merge_time[i] = 0;
    sort_time[i] = 0;
    scatter_time[i] = 0;
  }
  gather_time = 0.0;
  total_untrans = total_trans = total_merge = total_sort = total_scatter = 0.0;

  create_keys(np, localTemp, distrib);
  barrier();  
  start=get_seconds();
  bitonic_sort(np);
  finish=get_seconds();
  
  print_results(np);
  all_check_results(np, localData, 1);
}


/*
 * Procedure: print_results
 * Paramters: none
 */
void print_results(int np)
{
  int i;

  on_one {
    printf("Procs: %d Keys/PROC: %d  Total:  %.3lf s  Transpose+Gather: %.3lf Swap: %.3lf s  Untrans+Scatter: %.3lf Sort+Merge: %.3lf s Gather: %.3lf Scatter: %.3lf\n", 
           PROCS, np, finish - start, total_trans, total_merge, total_untrans, 
           total_sort, gather_time, total_scatter);
    printf("-----------------\n");
    printf("Components:\n");

    for (i = 1; merge_time[i] != 0.0; i++)
      printf(" merge[%d]: %.3lf us\n", i, merge_time[i] * 1e6);

    printf("-----------------\n");
    for (i = 1; trans_time[i] != 0.0; i++)
      printf(" trans[%d]: %.3lf us\n", i, trans_time[i] * 1e6);

    printf("-----------------\n");
    for (i = 0; sort_time[i] != 0.0; i++)
      printf(" sort[%d]: %.3lf us\n", i, sort_time[i] * 1e6);

    printf("-----------------\n");
    for (i = 1; untrans_time[i] != 0.0; i++)
      printf(" untrans[%d]: %.3lf us\n", i, untrans_time[i] * 1e6);

    printf("-----------------\n");
    for (i = 1; scatter_time[i] != 0.0; i++)
      printf(" scatter[%d]: %.3lf us\n", i, scatter_time[i] * 1e6);

    printf("\n\n");
  }

  barrier();
}


/*
 * Procedure: bitonic_sort
 * Parameters: np (input) the number of keys per processor
 *
 * Sorts the keys in localTemp and places them in localData.
 */
void   bitonic_sort(int np)
{
  int outerdim, offset;
  int *temp, p, i;

  /*
   * The first log np steps are replaced with a local sort of the keys
   * on each of the processors
   */
  offset = np / PROCS;
  sort_time[0] = get_seconds();
  bitonic_local_radix_sort(np, localTemp, localData);
  sort_time[0] = get_seconds() - sort_time[0];
  total_sort += sort_time[0];

  /* 
   * log PROCS merge stages are performed
   */
  for (outerdim = 1; outerdim <= logPROCS; outerdim ++, offset <<= 1){
    /*
     * The keys are transposed such that the keys are moved from a blocked
     * to a cyclic layout.  The first time through the loop we have to perform
     * a gather; this is not necessary for the other times because we perform
     * the gather in the last step of the loop: the merge.
     */
    if (outerdim == 1) {
      gather_time = get_seconds();
      gather(np, localTemp, localData);
      gather_time = get_seconds() - gather_time;

      trans_time[outerdim] = get_seconds();
      all_transpose(np, localData, localTemp);
      temp = localTemp;
      localTemp = localData;
      localData = temp;
    }
    else {
      trans_time[outerdim] = get_seconds();
      all_transpose(np, localTemp, localData);
    }
    trans_time[outerdim] = get_seconds() - trans_time[outerdim];
    total_trans += trans_time[outerdim];

    /*
     * Perform the operations of the butterfly 
     */
    merge_time[outerdim] = get_seconds();
    cyclic_merge(outerdim, np, localTemp, offset);
    barrier();
    merge_time[outerdim] = get_seconds() - merge_time[outerdim];
    total_merge += merge_time[outerdim];

    /*
     * Transpose and scatter the keys back to the blocked layout
     */
    untrans_time[outerdim] = get_seconds();
    all_transpose(np, localData, localTemp);
    untrans_time[outerdim] = get_seconds() - untrans_time[outerdim];
    total_untrans += untrans_time[outerdim];
    
    scatter_time[outerdim] = get_seconds();
    scatter(np, localTemp, localData);
    scatter_time[outerdim] = get_seconds() - scatter_time[outerdim];
    total_scatter += scatter_time[outerdim];

    /*
     * At this point, each of the processors has a bitonic sequence of keys; 
     * a sort is performed on the local keys that takes advantage of that
     * structure.  The gather required for the next iteration is folded
     * into the merge, except for on the last iteration.
     */
    sort_time[outerdim] = get_seconds();
    if (outerdim == logPROCS)
      one_merge(outerdim, np,localTemp, localData); 
    else
      one_merge_gather(outerdim, np, localTemp, localData); 
    barrier();
    sort_time[outerdim] = get_seconds() - sort_time[outerdim];
    total_sort += sort_time[outerdim];
  }
}

/* -------------------------------------------------------------------- 
 *
 * Routines to perform butterly merge stages with cyclic mapping
 *
 * -------------------------------------------------------------------- 
 */
/*
 * Procedure: cyclic_merge
 * Parameters:
 *   outerdim (input): the current merge stage
 *   np (input): number of keys per processor
 *   s (i/o): the keys
 *   offset (input): the size of the first group of keys that need to
 *                   be compared and swapped
 */
void cyclic_merge(unsigned outerdim, int np, int *s, int offset)
{

  int a0, a1, a2, a3;
  int a4, a5, a6, a7;
  int b0, b1, b2, b3;
  int b4, b5, b6, b7;
  int i, g, groups;
  int aptr, bptr;
  int innerdim, p;
  comp_t mode;

  /*
   * We need to perform outerdim merge steps.  
   * Each processor has groups collections of keys; the distance between
   * pairs of keys that are to be compared and swapped is offset.
   * The number of groups doubles on each merge step, while the offset is
   * halved.
   */
  groups = (np>>1)/offset;
  for (innerdim = outerdim-1; innerdim >= 0; innerdim--) {

    /*
     * To get better cache performance, if the two elements to be compared
     * fit on the same cache line, we do a different algorithm for the
     * remaining steps.
     */
    if (offset < 8) {
      cyclic_merge_cache(np, s, offset, groups, outerdim, innerdim);
      return;
    }
    /*
     * Each processor looks at each of its groups and determines if
     * it is to put the minimum or maximum key in the top position.
     */
    aptr = 0; bptr = offset; 
    for (g = 0; g < groups; g++) {
      mode = getMode(aptr/(np/PROCS), innerdim, outerdim);
      if (mode == COMP_MIN) {
        /*
         * We unroll the loop and look at groups of eight at a time so that 
         * we can read an entire cache line at once.  Otherwise, the two keys 
         * that are an "offset" apart often map to the same cache line and
         * we continually miss in the cache.
         */
        for (i = offset; i > 0; i-=8) {
          a0 = s[aptr]; a1 = s[aptr+1]; a2 = s[aptr+2]; a3 = s[aptr+3];
          a4 = s[aptr+4]; a5 = s[aptr+5]; a6 = s[aptr+6]; a7 = s[aptr+7];
          b0 = s[bptr]; b1 = s[bptr+1]; b2 = s[bptr+2]; b3 = s[bptr+3];
          b4 = s[bptr+4]; b5 = s[bptr+5]; b6 = s[bptr+6]; b7 = s[bptr+7];

          if (a0 > b0) { 
            s[aptr] = b0; s[bptr] = a0; 
          }
          if (a1 > b1) {
            s[aptr+1] = b1; s[bptr+1] = a1;
          }
          if (a2 > b2) {
            s[aptr+2] = b2; s[bptr+2] = a2;
          }
          if (a3 > b3) {
            s[aptr+3] = b3; s[bptr+3] = a3;
          }
          aptr+=4; bptr+=4; 
          if (a4 > b4) {
            s[aptr] = b4; s[bptr] = a4;
          }
          if (a5 > b5) {
            s[aptr+1] = b5; s[bptr+1] = a5;
          }
          if (a6 > b6) {
            s[aptr+2] = b6; s[bptr+2] = a6;
          }
          if (a7 > b7) {
            s[aptr+3] = b7; s[bptr+3] = a7;
          }
          aptr+=4; bptr+=4; 
        }
      }
      else {
        for (i = offset; i > 0; i-=8) {
          a0 = s[aptr]; a1 = s[aptr+1]; a2 = s[aptr+2]; a3 = s[aptr+3];
          a4 = s[aptr+4]; a5 = s[aptr+5]; a6 = s[aptr+6]; a7 = s[aptr+7];
          b0 = s[bptr]; b1 = s[bptr+1]; b2 = s[bptr+2]; b3 = s[bptr+3];
          b4 = s[bptr+4]; b5 = s[bptr+5]; b6 = s[bptr+6]; b7 = s[bptr+7];
          if (a0 < b0) {
            s[aptr] = b0; s[bptr] = a0;
          }
          if (a1 < b1) {
            s[aptr+1] = b1; s[bptr+1] = a1;
          }
          if (a2 < b2) {
            s[aptr+2] = b2; s[bptr+2] = a2;
          }
          if (a3 < b3) {
            s[aptr+3] = b3; s[bptr+3] = a3;
          }
          aptr+=4; bptr+=4; 
          if (a4 < b4) {
            s[aptr] = b4; s[bptr] = a4;
          }
          if (a5 < b5) {
            s[aptr+1] = b5; s[bptr+1] = a5;
          }
          if (a6 < b6) {
            s[aptr+2] = b6; s[bptr+2] = a6;
          }
          if (a7 < b7) {
            s[aptr+3] = b7; s[bptr+3] = a7;
          }
          aptr+=4; bptr+=4; 
        }
      }
      aptr += offset; bptr += offset; 
    }
    offset >>= 1; groups <<= 1;
  }
}

/*
 * Procedure: cyclic_merge_cache
 * Parameters: 
 *   np (input): number of keys per processor
 *   s (i/o): the keys
 *   offset (input): the distance between keys to be compared; guaranteed to
 *                   be less than eight
 *   groups (input): the number of sets of keys that are at a distance of offset
 *                   apart
 *   outerdim (input): the current merge stage
 *   innerdim (input): the current merge step of that merge stage
 */
void cyclic_merge_cache(int np, int *s, int offset, int groups, int outerdim, 
            int innerdim)
{
  int a, b, g;
  comp_t mode;
  int aptr, bptr, i, inner;
  int g2, gr, off;

  /*
   * Do each of the merge steps for this group before proceeding to the
   * next group; we know that the keys to be compared are on the same
   * cache line
   */
  for (g = 0; g < groups; g++) {
    off = offset;
    gr = 1;
    for (inner = innerdim; inner >= 0; inner--) {
      /*
       * The number of groups doubles as we progress through the merge
       * steps
       */
      for (g2 = 0; g2 < gr; g2++) {
        aptr = g2*(off<<1) + g*(offset<<1); 
        bptr = aptr + off; 
        mode = getMode(aptr/(np/PROCS), inner, outerdim);
        if (mode == COMP_MIN) {
          for (i = off; i > 0; i--) {
            if ((a = s[aptr]) > (b = s[bptr])) {
              s[aptr] = b; 
              s[bptr] = a;
            }
            aptr++; bptr++; 
          }
        }
        else {
          for (i = off; i > 0; i--) {
            if ((a = s[aptr]) < (b = s[bptr])) {
              s[aptr] = b; 
              s[bptr] = a;
            }
            aptr++; bptr++; 
          } 
        }
      } 
      off >>= 1; gr <<= 1;
    }
  }
}


/* -------------------------------------------------------------------- 
 *
 * Routines to sort a bitonic sequence.  With and witout gathering.
 *
 * -------------------------------------------------------------------- 
 */
/*
 * Procedure: one_merge_gather
 * Parameters:
 *   outerdim (input): the merge stage
 *   np (input): the number of keys per processor
 *   src (input): src buffer of np keys -- a bitonic sequence
 *   dest (input): dest buffer -- a sorted sequence, but already gathered
 *                 for the transpose
 */
void one_merge_gather(int outerdim, int np, int *src, int *dest)        
{
  /*
   * Determine if we are to sort up or down
   */
  if (MYPROC & (1 << outerdim)) bmerge_up_gather(np, src, dest);
  else bmerge_down_gather(np, src, dest);
}


/* 
 * Procedure: bmerge_down_gather
 * Parameters: 
 *   np (input): keys per processor
 *   srcBuf (input): bitonic sequence of keys
 *   destBuf (input): sequence of keys sorted decreasingly, but
 *                    permuted for transpose
 */
void bmerge_down_gather(int np, int *srcBuf, int *destBuf)
{
  int m, n, k, min, sort_miss;
  int p, band = np/PROCS;
  int offset;

  min = find_bmin(np, srcBuf);
  if (min == -1) return;
  m = min;
  n = (min + 1) % np;

  /* 
   * Fill in destination array starting with smallest element.
   * Spread the keys cyclically through the buffer, but beginning
   * with the last key for the last processor
   */
  for (k = band-1; k >= 0; k--) {
    for (offset = (PROCS-1)*band + k, p = PROCS-1; p >= 0; p--, offset -= band) {
      if (srcBuf[m] >= srcBuf[n]) {
        destBuf[offset] = srcBuf[n];
        n = (n + 1) % np;
      }
      else {
        destBuf[offset] = srcBuf[m];
        m =  (m == 0) ? (np-1) : (m-1);
      }
    }
  }
}

/* 
 * Procedure: bmerge_up_gather
 * Parameters: 
 *   np (input): keys per processor
 *   srcBuf (input): bitonic sequence of keys
 *   destBuf (input): sequence of keys sorted increasingly, but
 *                    permuted for transpose
 */
void bmerge_up_gather(int np, int *srcBuf, int *destBuf)
{
  int m, n, k, min, sort_miss;
  int p, band = np/PROCS;
  int offset;

  /*
   * Fill in destination array, starting with smallest element.
   */
  min = find_bmin(np,srcBuf);
  if (min == -1) 
    /* The list is already sorted */
    return;
  m = min;
  n = (min + 1) % np;

  /*
   * Each processor receives band keys from this processor; spread
   * the keys cyclically across processors
   */
  for (k = 0; k < band; k++) {
    for (offset = k, p = 0; p < PROCS; p++, offset += band) {
      /*
       * Put the minimum key into the next cyclic position in the buffer.
       * We only have to choose between the two keys on either end of the 
       * minimum key.  
       * The 'n' index is incremented when chosen; the 'm' index is
       * decremented, taking care to wrap around the edges of the srcbuf.
       */
      if (srcBuf[m] >= srcBuf[n]) {
        destBuf[offset] = srcBuf[n];
        n = (n + 1) % np;
      }
      else {
        destBuf[offset] = srcBuf[m];
        m =  (m == 0) ? (np-1) : (m-1);
      }
    }
  }
}


/*
 * Procedure: one_merge
 * Parameters:
 *   outerdim (input): the merge stage
 *   np (input): the number of keys per processor
 *   src (input): src buffer of np keys -- a bitonic sequence
 *   dest (input): dest buffer -- a sorted sequence
 */
void one_merge(int outerdim, int np, int *src, int *dest)       
{
  if (MYPROC & (1 << outerdim)) bmerge_up(np, src, dest);
  else bmerge_down(np, src, dest);
}

    
/* 
 * Procedure: bmerge_down
 * Parameters: 
 *   np (input): keys per processor
 *   srcBuf (input): bitonic sequence of keys
 *   destBuf (input): sequence of keys sorted decreasingly
 */
void bmerge_down(int np, int *srcBuf, int *destBuf)
{
  int m, n, k, min;

  min = find_bmin(np, srcBuf);
  if (min == -1) return;

  /* fill in destination array from top starting with smallest element */
  m = min;
  n = (min + 1) % np;
  for (k = np-1; k >= 0; k--) {
    if (srcBuf[m] >= srcBuf[n]) {
      destBuf[k] = srcBuf[n];
      n = (n + 1) % np;
    }
    else {
      destBuf[k] = srcBuf[m];
      m =  (m == 0) ? (np-1) : (m-1);
    }
  }
}


/* 
 * Procedure: bmerge_up
 * Parameters: 
 *   np (input): keys per processor
 *   srcBuf (input): bitonic sequence of keys
 *   destBuf (input): sequence of keys sorted increasingly
 */
void bmerge_up(int np, int *srcBuf, int *destBuf)
{
  int m, n, k, min;

  min = find_bmin(np, srcBuf);
  if (min == -1) return;

  /* fill in destination array, starting with smallest element */
  m = min;
  n = (min + 1) % np;
  for (k = 0; k < np; k++) {
    if (srcBuf[m] >= srcBuf[n]) {
      destBuf[k] = srcBuf[n];
      n = (n + 1) % np;
    }
    else {
      destBuf[k] = srcBuf[m];
      m =  (m == 0) ? (np-1) : (m-1);
    }
  }
}


/*
 * Procedure: getMode
 * Parameters: 
 *   iproc (input): the node number in the butterfly corresponding to the
 *          current key
 *   innerdim (input): the merge step
 *   outerdim (intput): the merge stage
 *
 * Returns:
 *   If the iproc-th node should contain the minimum or the maximum key after
 *   the compare and swap.
 */
comp_t inline getMode(unsigned iproc, unsigned innerdim, unsigned outerdim)
{
  /* xor of bit (innerdim) of iproc and bit outerdim of iproc */
  unsigned int innerbit, outerbit;

  innerbit = iproc & (1 << innerdim);
  outerbit = iproc & (1 << outerdim);

  if((innerbit || outerbit) && !(innerbit && outerbit)) return COMP_MIN;
  return COMP_MAX;
}


/* -------------------------------------------------------------------- 
 *
 * Routines to find the minimum of a bitonic sequence
 *
 * -------------------------------------------------------------------- 
 */

/* 
 * Procedure: find_bmin
 * Parameters:
 *   np (input): keys per processor
 *   srcBuf (input): bitonic sequence of keysK
 * Returns:
 *   The index of the key which is the minimum, or -1 if all of the
 *   keys are indentical.
 */
int inline find_bmin(int np, int *srcBuf)
{
  /*
   * After determining the shape of the bitonic sequence, it finds
   * the minimum key, depending upon the shape.
   */
  switch (which_bitonic(np, srcBuf)) {
  case SINGLE_DOWN: 
    return find_bmin_single_down(0, np-1, srcBuf);
  case SINGLE_UP: 
    return find_bmin_single_up(0, np-1, srcBuf);
  case DOUBLE_UP: 
    return find_bmin_double_up(0, np-1, srcBuf);
  case DOUBLE_DOWN: 
    return find_bmin_double_down(0, np-1, srcBuf);
  case IDENTICAL:
    return -1;
  case SLOPE_UP:
    return 0;
  case SLOPE_DOWN:
    return np-1;
  }
}


/*
 * Procedure: which_bitonic
 * Parameters: 
 *   np (input): keys per processor
 *   s (input): keys
 * Returns:
 *   The shape of the bitonic sequence.  There are these options:
 *
 *  IDENTICAL  SLOPE_UP SLOPE_DOWN  SINGLE_UP  SINGLE_DOWN DOUBLE_UP DOUBLE_DOWN
 */
int inline which_bitonic(int np, int *s)
{
  int a, b, c, d;
  int i;

  a = s[0];
  b = s[np-1];

  /* 
   * Get the first key on the left that is different from a into c.
   * Put sentry values at ends so don't have to have two
   * conditionals in loops.  
   */
  s[np-1] = 0x80000000; 
  for (i = 1; (c = s[i]) == a; i++);
  /* Remember to replace original key when done! */
  s[np-1] = b;
  /* If all keys are identical, there isn't much else to do */
  if (i == np-1) {
    if (c == b) return IDENTICAL;
    /* 
     * Be careful of the case where the last key is different;
     */
    if (a > b) return SLOPE_DOWN;
    else return SLOPE_UP;
  }

  /*
   * Find the first key from the right that is different from b
   */
  s[0] = 0x80000000;    
  for (i = np-2; (d = s[i]) == b; i--);
  s[0] = a;
  /* If all keys are identical, there isn't much else to do */
  if (i == 0) {
    if (a == d) return IDENTICAL;
    /* 
     * Be careful of the case where the last key is different;
     */
    if (a > b) return SLOPE_DOWN;
    else return SLOPE_UP;
  }

  if (a > c) {
    if (b > d) return SINGLE_DOWN;
    else 
      if (c < d) return DOUBLE_DOWN;
      else return SLOPE_DOWN;
  }
  else {
    if (b < d) return SINGLE_UP;
    else 
      if (c > d) return DOUBLE_UP;
      else return SLOPE_UP;
  }
}


/*
 * Procedure: find_bmin_single_up
 * Parameters:   
 *   first (input): the first element of s that is relevant
 *   last (input): the last element of s
 *   s (input): s[f]...s[l] are a bitonic sequence with shape DOUBLE_UP
 */
int find_bmin_single_up(int first, int last, int *s)
{
  if (s[first] <= s[last]) return first;
  return last;
}

/*
 * Procedure: find_bmin_single_down
 * Parameters:
 *   first (input): the first element of s that is relevant
 *   last (input): the last element of s
 *   s (input): s[f]...s[l] are a bitonic sequence with shape DOUBLE_UP
 */
int find_bmin_single_down(int first, int last, int *s)
{

  int mid = (first + last)>>1;

  switch (which_dir_point(first, DOWN_POINT, last, UP_POINT, mid, s)) {
  case UP_POINT: 
    return find_bmin_single_down(first, mid, s);
  case DOWN_POINT: 
    return find_bmin_single_down(mid, last, s);
  case DOWN_MID:
  case IDENTICAL:
    return mid;
  }
}


/*
 * Procedure: find_bmin_double_up
 * Parameters:
 *   f (input): the first element of s that is relevant
 *   l (input): the last element of s
 *   s (input): s[f]...s[l] are a bitonic sequence with shape DOUBLE_UP
 * Returns:
 *  The minimum key of s between f and l.
 */
int find_bmin_double_up(int f, int l, int *s) 
{
  int m = (f + l) >> 1;

  switch (which_dir_point(f, UP_POINT, l, UP_POINT, m, s)) {
  case DOWN_POINT:
  case UP_MID:
    return find_bmin_single_down(m, l, s);
  case DOWN_MID:
    return m;
  case UP_POINT:
    if (s[m] < s[l]) return find_bmin_double_up(f, m, s);
    return find_bmin_double_up(m, l, s);
  default:
    printf("ERROR: Unknown type\n.");
  }
}


/*
 * Procedure: find_bmin_double_down
 * Parameters:
 *   f (input): the first element of s that is relevant
 *   l (input): the last element of s
 *   s (input): s[f]...s[l] are a bitonic sequence with shape DOUBLE_DOWN
 * Returns:
 *  The minimum key of s between f and l.
 */
int find_bmin_double_down(int f, int l, int *s) 
{
  int m = (f + l) >> 1;

  /*
   * Recursively look at keys 
   */
  switch (which_dir_point(f, DOWN_POINT, l, DOWN_POINT, m, s)) {
  case UP_POINT:
  case UP_MID:
    return find_bmin_single_down(f, m, s);
  case DOWN_MID:
    return m;
  case DOWN_POINT:
    if (s[m] < s[f]) return find_bmin_double_down(m, l, s);
    return find_bmin_double_down(f, m, s);
  }
}


/*
 * Procedure: which_dir_point
 * Parameters: 
 *   f (input): the index of the left-most key
 *   f_dir (input): the type of that left-most key
 *   l (input): the index of the right-most key
 *   l_dir (input): the type of that right-most key
 *   m (input): the index of the middle key
 *   s (input): the keys
 * Returns:
 *   The type of the middle key.
 *   UP_POINT: the middle key is on an upwards slope
 *   DOWN_POINT: the middle key is on a downwards slope
 *   UP_MID: the middle key is at a peak
 *   DOWN_MID: the middle key is at a valley
 *   IDENTICAL: all keys are identical
 * If there are
 */
int inline which_dir_point(int f, int f_dir, int l, int l_dir, int m, int *s)
{
  int i;
  int a, b, c;

  a = b = c = s[m];

  /* Find the first key different from s[m] to the right */
  for (i = m+1; i <= l; i++)
    if ((c = s[i]) != b) break;

  /* Find the first key different from s[m] to the left */
  for (i = m-1; i >= f; i--)
    if ((a = s[i]) != b) break;

  if (a < b && b < c) return UP_POINT;
  if (a > b && b > c) return DOWN_POINT;
  if (a < b && c < b) return UP_MID;
  if (a > b && c > b) return DOWN_MID;

  if (a == b)
    if (c == b) return IDENTICAL;
    else return f_dir;

  /* c == b */
  return l_dir;
}


/* -------------------------------------------------------------------- 
 *
 * Routines to perform local sort on arbitrary keys.
 *
 * -------------------------------------------------------------------- 
 */
/*
 * Procedure: bitonic_local_radix_sort
 * Parameters: 
 *   np (input): the number of keys to sort
 *   src (input): the list of np keys to be sorted
 *   dest (output): the list of np keys, sorted in either
 *         increasing or decreasing order, depending upon
 *         the processor number
 */
void bitonic_local_radix_sort(int np, int *src, int *dest)      
{
  if (MYPROC & 1) local_radix_sort(np, src, dest);
  else local_radix_sort_down(np, src, dest);
}
