/*                      
 * column.sc - Split-C implementation of column sort.
 */
#include <stdio.h>
#include <split-c/split-c.h>
#include "sort.h"
#include "local_radix.h"
#include "transpose.h"
#include "column.h"

/* Statically allocate NP_MAX keys of buffer space on each processor */
#define NP_MAX (1024*1024)
int matrix1[PROCS]::[NP_MAX];
int matrix2[PROCS]::[NP_MAX];

/* Pointers to the local buffer space */
int *col1, *col2;

/* The time of the entire algorithm and the nine phases */
double t[10];

void splitc_main(int argc, char *argv[])
{
  int c;
  extern char *optarg;
  int distrib = 0, np = 16 * 1024;

  /* Get options from the user */
  while ((c = getopt(argc, argv, "r:d:")) != -1) {
    switch(c) {
    case 'r':  np = atoi(optarg);  np *= 1024; break;
    case 'd':  distrib = atoi(optarg); break;
    }
  }

  /* Check options -- number of keys is constrained. */
  if (np % PROCS) {
    on_one {
      printf("ERROR: np (%d) must be evenly divisible by the number of processors.\n", np);
      exit(-1);
    }
  }
  if (np < 2 * (PROCS-1) * (PROCS-1)) {
    on_one {
      printf("ERROR: np(%d) must be >= 2 * (PROCS-1)^2 (%d).\n", np, 
             2 * (PROCS-1) * (PROCS-1));
      exit(-1);
    }
  }
  if (np > NP_MAX) 
    on_one {
      printf("ERROR: np (%d) must be <= %d\n", np, NP_MAX);
      exit(-1);
    }

  on_one  {
    printf("\nSorting %d keys on %d processors (%d keys/proc).\n", np*PROCS, PROCS, np);
    printf("Key distribution: %d\n", distrib);
  }

  col1 = (int *)matrix1;
  col2 = (int *)matrix2;
  create_keys(np, col1, distrib);
  barrier();


  t[0] = get_seconds();
  t[1] = get_seconds();
  local_radix_sort(np, col1, col2);
  barrier();
  t[1] = get_seconds()-t[1];

  t[2] = get_seconds();
  gather(np, col1, col2);
  barrier();
  t[2] = get_seconds()-t[2];

  t[3] = get_seconds();
  all_transpose(np, col2, col1);
  barrier();
  t[3] = get_seconds()-t[3];

  t[4] = get_seconds();
  local_radix_sort(np, col2, col1);
  barrier();
  t[4] = get_seconds() - t[4];

  t[5] = get_seconds();
  all_transpose(np, col2, col1);
  barrier();
  t[5] = get_seconds() - t[5];

  t[6] = get_seconds();
  local_radix_sort(np, col2, col1);
  barrier();
  t[6] = get_seconds() - t[6];

  t[7] = get_seconds();
  all_shift(np, col2, col1);
  barrier();
  t[7] = get_seconds() - t[7];

  t[8] = get_seconds();
  if (MYPROC) merge(np, col2, col1);
  else {
    int i;
    int a0, a1, a2, a3, a4, a5, a6, a7;
    for (i = 0; i < np; i += 8) {
      a0 = col2[i];   a1 = col2[i+1]; a2 = col2[i+2]; a3 = col2[i+3];
      a4 = col2[i+4]; a5 = col2[i+5]; a6 = col2[i+6]; a7 = col2[i+7];
      col1[i] = a0; col1[i+1] = a1; col1[i+2] = a2; col1[i+3] = a3;
      col1[i+4] = a4; col1[i+5] = a5; col1[i+6] = a6; col1[i+7] = a7;
    }
  }
  barrier();
  t[8] = get_seconds() - t[8];

  t[9] = get_seconds();
  all_unshift(np, col2, col1);
  barrier();
  t[9] = get_seconds() - t[9];

  t[0] = get_seconds()-t[0];
  print_results(np);
  all_check_results(np, col2, 0);
}


/*
 * Procedure: all_shift
 * Parameters: 
 *   s (input): the list of n keys to be shifted
 *   d (output): the list of keys 'shifted' 
 *
 * Note: The formal description of column sort states that the
 *       first half of s should be shifted to the second half of d
 *       and that the second half of s becomes the first half d, but
 *       because the next step is to merge the two halves, the ordering
 *       is inconsequential.
 */
void all_shift(int np, int *d, int *s)
{
  int *global dst;
  int i, half_keys;
  int a0, a1, a2, a3, a4, a5, a6, a7;

  /*
   * Do the copy of the local data: the first half of d contains the
   * first half of s on this processor.  The code is unrolled for cache lines
   * that contain 32 bytes.
   */
  half_keys = np >> 1;
  for (i = 0; i < half_keys; i += 8) {
    a0 = s[i];   a1 = s[i+1]; a2 = s[i+2]; a3 = s[i+3];
    a4 = s[i+4]; a5 = s[i+5]; a6 = s[i+6]; a7 = s[i+7];
    d[i] = a0; d[i+1] = a1; d[i+2] = a2; d[i+3] = a3;
    d[i+4] = a4; d[i+5] = a5; d[i+6] = a6; d[i+7] = a7;
  }

  barrier();
  /*
   * Store the second half of the keys to the the second half of the 
   * next processor and then wait until I have received my keys.
   */
  dst = toglobal(MYPROC != PROCS-1 ? MYPROC+1 : 0, d + half_keys);
  bulk_store(dst, s + half_keys, sizeof(int)*half_keys);
  store_sync(sizeof(int)*half_keys);
}


/*
 * Procedure: all_unshift
 * Parameters: s (input) list of n keys
 *             d (output) list of n keys 'unshifted'.
 */
void all_unshift(int np, int *d, int *s)
{
  int *global dst;
  int i, half_keys, j;
  int a0, a1, a2, a3, a4, a5, a6, a7;

  half_keys = np >> 1;
  barrier();

  /*
   * Processor 0 keeps the first half the keys in place and stores its
   * second half as the second half on the last processor
   */
  if (!MYPROC) {
    for (i = 0; i < half_keys; i += 8) {
      a0 = s[i];   a1 = s[i+1]; a2 = s[i+2]; a3 = s[i+3];
      a4 = s[i+4]; a5 = s[i+5]; a6 = s[i+6]; a7 = s[i+7];
      d[i] = a0; d[i+1] = a1; d[i+2] = a2; d[i+3] = a3;
      d[i+4] = a4; d[i+5] = a5; d[i+6] = a6; d[i+7] = a7;
    }
    dst = toglobal(PROCS-1, d + half_keys);
    bulk_store(dst, s + half_keys, sizeof(int)*half_keys);

  }
  /*
   * All other processors construct d such that the first half of d
   * contains the second half of s on this processor and the second 
   * half of d contains the first half of s on the next processor.  
   */
  else {
    for (i = 0, j = half_keys; i < half_keys; i += 8, j += 8) {
      a0 = s[j];   a1 = s[j+1]; a2 = s[j+2]; a3 = s[j+3];
      a4 = s[j+4]; a5 = s[j+5]; a6 = s[j+6]; a7 = s[j+7];
      d[i] = a0; d[i+1] = a1; d[i+2] = a2; d[i+3] = a3;
      d[i+4] = a4; d[i+5] = a5; d[i+6] = a6; d[i+7] = a7;
    }
    dst = toglobal(MYPROC-1, d + half_keys);
    bulk_store(dst, s, sizeof(int)*half_keys);
  }

  store_sync(sizeof(int)*half_keys);
}


/*
 * Procedure: merge
 * Parameters: 
 *   np:      (input) the number of keys per processor
 *   srcBuf:  (input) a list of n keys, such that the first np/2 elements are 
 *            sorted, as are the last np/2 elements
 *   destbuf: (output) the sorted list of np keys
 *
 */
void merge(int np, int *srcBuf, int *destBuf)
{
  int j, end1, end2;
  int i, k;

  j = 0;
  end1 = i = j + (np >>1);
  end2 = end1 + (np >>1);

  /* 
   * Pick lesser of the elements in each of the two lists and put it
   * in the destination buffer.  You break out of the loop as soon as
   * you reach the end of one of the two buffers being merged. 
   */
  for (k = 0; ; ) {

    if (srcBuf[j] <= srcBuf[i]) {
      destBuf[k++] = srcBuf[j++];
      if ( j == end1 ) break;
    }
    else {
      destBuf[k++] = srcBuf[i++];
      if ( i == end2 ) break;
    }
  }

  /* 
   * Check whether i or j got over first and then just copy
   * the elements from the other buffer
   */
  if ( j == end1 )
    for ( ; i < end2 ; )
      destBuf[k++] = srcBuf[i++];
  else
    for ( ; j < end1 ; )
      destBuf[k++] = srcBuf[j++];
}


/*
 * Procedure: print_results
 * Parameters: 
 *   np (input) keys per processor 
 * 
 * Prints the average across processors of the time per phase of the algorithm
 * in seconds.
 */
void print_results(int np)
{
  int i;
  double max, min, sum;

  on_one printf("Procs: %d   Keys/PROCS: %d   ", PROCS, np);
  for (i = 0; i < 10; i++) {
    
    sum = all_reduce_to_one_dadd(t[i]);
    on_one  {
      switch (i) {
      case 0: printf("Total --"); break;
      case 1: printf("Sort --"); break;
      case 2: printf("Gather --"); break;
      case 3: printf("Trans --"); break;
      case 4: printf("Sort --"); break;
      case 5: printf("Untrans --"); break;
      case 6: printf("Sort --"); break;
      case 7: printf("Shift --"); break;
      case 8: printf("Merge --"); break;
      case 9: printf("Unshift --"); break;
      }
      printf(" %.3lf s   ", sum/PROCS);
    }
  } 
  on_one printf("\n");
}





  
      

  
















