/* 
 * local_radix.sc - Local radix sort
 */
#include <stdio.h>
#include <split-c/split-c.h>
#include <split-c/com.h>
#include "local_radix.h"

/*
 * An 11-bit radix is optimal for input sizes between 16K and 1M
 */
#define RADIX_11         2048
#define RADIX11_MIN_1    2047
#define RADIX_10         1024

/*
 * 8-bit radix for smaller numbers of keys
 */
#define RADIX_8          256
#define RADIX8_MIN_1     255

/*
 * The buckets which count the number of occurences of each digit in
 * the list of keys. 
 */
static unsigned int   buck[RADIX_11];


/*
 * Procedure: local_radix_sort
 * Parameters:
 *   np (input): number of keys to sort 
 *   c (input): the list of np keys to be sorted
 *   t(output): the increasingly sorted list of keys
 */
void local_radix_sort(int np, int *c, int *t)
{
  unsigned int *x,mask,unshift;
  int i, j, sum, temp;
  int pass;

  mask = RADIX11_MIN_1;
  unshift = 0;
  for (pass =0; pass < 2 ; pass++) {    
    
    /*
     * Initialize the bucket counts
     */
    for (j = 0; j < RADIX_11; j++)
      buck[j] = 0;
      
    /*
     * Set buck[i] equal to the number of occurences of
     * digit i in the list of keys
     */
    for (j = 0; j < np; j++) {
      buck[(c[j] & mask) >> unshift]++;
    }

    /*
     * Scan across the buckets, settin buck[i] equal to the 
     * sum of buck[0..i-1]
     */
    sum = 0;
    for (j = 0; j < RADIX_11; j++) {
      temp = buck[j];
      buck[j] = sum;
      sum += temp;
    }
    
    /*
     * Permute a key with digit value i according to its rank, which
     * is designated by the buck[i].
     */
    for (j = 0; j < np; j++) {
      sum = buck[((temp=c[j]) & mask) >> unshift]++;
      t[sum] = temp;
    }

    /*
     * Now look at the next higher digit, using as our src buffer
     * what we just had as our destination buffer.
     */
    x = t;
    t = c;
    c = x;
    mask = mask <<  11;
    unshift += 11;
  }

  /*
   * Perform the same operations on the last 10 bits of the
   * keys
   */
  for (j = 0; j < RADIX_10; j++)
    buck[j] = 0;
      
  for (j = 0; j < np; j++) {
    buck[ (((c[j])&mask)>> unshift)]++;
  }

  sum = 0;
  for (j = 0; j < RADIX_10; j++) {
    temp = buck[j];
    buck[j] = sum;
    sum += temp;
  }
      
  for (j = 0; j < np; j++) {
    sum = buck[ (((temp=c[j])&mask)>> unshift)]++;
    t[sum] = temp;
  }
}


/*
 * Procedure: local_radix_sort_down
 * Parameters:
 *   np (input): number of keys to sort 
 *   c (input): the list of np keys to be sorted
 *   t(output): the decreasingly sorted list of keys
 */
void local_radix_sort_down(int np, int *c, int *t)
{
  unsigned int *x,mask,unshift;
  int i, j, sum, temp;
  int pass;
 
  mask = RADIX11_MIN_1;
  unshift = 0;

  for (pass = 0; pass < 2; pass++) {    

    for (j = 0; j < RADIX_11; j++)
      buck[j] = 0;
      
    for (j = 0; j < np; j++) {
      buck[(c[j]&mask) >> unshift]++;
    }

    sum = 0;
    for (j = 0; j < RADIX_11; j++) {
      temp = buck[j];
      buck[j] = sum;
      sum += temp;
    }
      
    for (j = 0; j < np; j++) {
      sum = buck[((temp=c[j]) & mask)>> unshift]++;
      t[sum] = temp;
    }

    x = t;
    t = c;
    c = x;
    mask = mask <<  11;
    unshift += 11;
  }

  for (j = 0; j < RADIX_10; j++)
    buck[j] = 0;
  
  for (j = 0; j < np; j++) {
    buck[(c[j] & mask) >> unshift]++;
  }

  sum = 0;
  for (j = 0; j < RADIX_10; j++) {
    temp = buck[j];
    buck[j] = sum;
    sum += temp;
  }
      
  for (j = 0; j < np; j++) {
    sum = buck[((temp=c[j]) & mask) >> unshift]++;
    t[(np-1) - sum] = temp;
  }
}


/*
 * Procedure: local_radix_sort_8bit
 * Parameters:
 *   np (input): number of keys to sort 
 *   c (input): the list of np keys to be sorted
 *    (output): the increasingly sorted list of keys
 *   t        : temporary buffer space
 */
void local_radix_sort_8bit(int np, int *c, int *t)
{
  unsigned int *x,mask,unshift;
  int i, j, sum, temp;
  int pass;

  mask = RADIX8_MIN_1;
  unshift = 0;
  for (pass = 0; pass < 4; pass++) {    
    
    /*
     * Initialize the bucket counts
     */
    for (j = 0; j < RADIX_8; j++)
      buck[j] = 0;
      
    /*
     * Set buck[i] equal to the number of occurences of
     * digit i in the list of keys
     */
    for (j = 0; j < np; j++) {
      buck[(c[j] & mask) >> unshift]++;
    }

    /*
     * Scan across the buckets, settin buck[i] equal to the 
     * sum of buck[0..i-1]
     */
    sum = 0;
    for (j = 0; j < RADIX_8; j++) {
      temp = buck[j];
      buck[j] = sum;
      sum += temp;
    }
    
    /*
     * Permute a key with digit value i according to its rank, which
     * is designated by the buck[i].
     */
    for (j = 0; j < np; j++) {
      sum = buck[((temp=c[j]) & mask) >> unshift]++;
      t[sum] = temp;
    }

    /*
     * Now look at the next higher digit, using as our src buffer
     * what we just had as our destination buffer.
     */
    x = t;
    t = c;
    c = x;
    mask = mask <<  8;
    unshift += 8;
  }
}
