/*                                  
 * radix.sc - Split-C radix sort program
 */
#include <split-c/split-c.h>
#include <split-c/control.h>
#include <split-c/com.h>
#include "sort.h"

/* 
 * Two key lists are used--on each pass the keys are moved from one to
 * the other, sorting a specific digit (least significant first). 
 */
#define NP_MAX (1024*1024)
unsigned g_keys0[PROCS]::[NP_MAX];
unsigned g_keys1[PROCS]::[NP_MAX]; 
unsigned *keys;

/*
 * DIGITS_PER_KEY * BITS_PER_DIGIT == number of bits in key
 */
#define DIGITS_PER_KEY 2      /* digits in a key (round up) */
#define BITS_PER_DIGIT 16           

/* 
 * Number of buckets == 2^(BITS_PER_DIGIT) 
 */
#define RADIX          65536        
#define RADIX_MASK     (RADIX-1)

/* 
 * A list of buckets provides a count of how many keys on a given
 * processor have a certain value of the digit being sorted. 
 */
unsigned g_buckets[PROCS]::[RADIX];
unsigned *buckets;

/* 
 * A list of global offsets into the key list allows each processor to
 * write its keys into their new locations. 
 */
int g_scan_buffer[PROCS]::[RADIX];
int *scan_buffer;
int g_scan_result[PROCS]::[RADIX];
int *scan_result;


/* 
 * Timing data 
 */
double time_start, time_finish;
double total_fill, total_hist,total_scan, total_coallesce;
double pass_time[DIGITS_PER_KEY][5];

/*  
 * Procedure: init_buckets
 * Parameters: none
 *
 * Buckets are initialized to 0.
 */
void inline init_buckets()
{
  int i;
    
  for (i = 0; i < RADIX; i++)
    buckets[i] = 0;
}

/*
 * Procedure: histogram
 * Parameters: 
 *   np (input): number of keys per processor
 *   pass (input): the current pass, beginning with 0.
 *   keys (input): local list of keys
 *   
 * Bucket[i] is set to the number of occurrences of the
 * digit i in the list of keys.
 */
void inline histogram(int np, int pass, int *keys) 
{
  int i, shift = pass*BITS_PER_DIGIT;

  for (i = 0; i < np; i++)
    buckets[(keys[i] >> shift) & RADIX_MASK]++;
}

/* 
 * Procedure: all_coalesce 
 * Parameters:
 *   np (input): keys per processor
 *   log_keys (input): log(np)
 *   pass (input): the pass number, beginning with 0
 *   keys (input): the local list of keys that are ranked in
 *                 the buckets array
 *   g_keys (output): the keys after they have been moved to their
 *                    position across processors
 */
void all_coalesce(int np, int log_keys, int pass, unsigned *keys, 
            unsigned (* spread g_keys)[NP_MAX])
{
  int i, shift = pass*BITS_PER_DIGIT;
  int key_bucket, new_row, new_column;
  unsigned *global g_ptr;
  int np_mask = np-1;
  int k;

  for (i = np-1; i >= 0; i--) {
    key_bucket = ((k = keys[i]) >> shift) & RADIX_MASK; 
    buckets[key_bucket]--; 
    new_row = buckets[key_bucket] >> log_keys;
    new_column = buckets[key_bucket] & np_mask;
    g_ptr = g_keys[new_row];
    g_ptr += new_column; 
    *g_ptr :- k; 
  }

  all_store_sync();
}

/*
 * Procedure: all_scan_buckets
 * Parameters: none
 * 
 * On entry, buckets contains the local histogram on each processor. 
 * On exit, buckets[p][i] contains:
 *    sum buckets[0..PROCS-1][0..(i-1)] +
 *    sum buckets[0..p][i]
 */ 
void all_scan_buckets()
{
  int i,b,proc;
  int sum =0;
  int *global neighbor;
  int j, k;


  /* 
   * Use -1 to designate that a position is empty; works
   * only because keys are only 31 bits.
   */
  for(i=0; i < RADIX ; i++) {   
    scan_buffer[i] = -1;
    scan_result[i] = -1;
  }
  neighbor = toglobal((MYPROC+1) % PROCS, scan_buffer);
  barrier(); 

  /*
   * Processor 0 sends each of its RADIX buckets to successive positions
   * on the neighboring processor.  It waits an extra o + tadd between
   * sends so that it doesn't flood the destination processor faster
   * than it can retrieve messages from the network and forward them.
   */
  if (MYPROC == 0) {
    for (b = 0; b < RADIX; b++) {
      *neighbor++ :- buckets[b];
      /* 
       * Each iteration through the loop takes about 0.1 usec
       */
      for (j = 36; j; j--) 
        k += j;
    }
  }
  /*
   * The other processors wait until the value has arrived (i.e., it
   * is not -1); it then adds the received value to its local bucket
   * value and, except for the last processor, then forwards the
   * result to its neigbor. The last processor stores the result in
   * the scan_result array.
   */
  else {
    if (MYPROC != (PROCS-1)) {
      for (b = 0; b < RADIX ; b++) {
        /* spin waiting for store to complete */
        while (scan_buffer[b] == -1) splitc_poll();           
        *neighbor++ :- scan_buffer[b] + buckets[b];
      }
    }
    else {
      for (b = 0; b < RADIX ; b++) {
        /* spin waiting for store to complete */
        while (scan_buffer[b] == -1) splitc_poll();           
        scan_result[b] = scan_buffer[b] + buckets[b];
      }
    }
  }

  barrier();

  /*
   * The sum of each row of buckets is now broadcast
   * to all of the processors in the scan_result array
   */
  neighbor = toglobal((MYPROC+1) % PROCS, scan_result);

  /*
   * The last processor stores each of the buckets to 
   * processor 0; it waits after each send to avoid
   * flooding processor 0.
   */
  if (MYPROC == (PROCS-1)) {
    for (b = 0; b < RADIX; b++) {
      *neighbor++ :- scan_result[b];
      for (j = 24; j; j--) 
        k += j;
    } 
  }
  else {
    if (MYPROC < (PROCS-2)) {
      for (b = 0; b < RADIX; b++) {
        while (scan_result[b] == -1) splitc_poll();
        *neighbor++ :- scan_result[b];
      }
    }
    /*
     * The last processor in the pipeline does not
     * forward on the value
     */
    else {
      for (b = 0;b < RADIX; b++) {
        while (scan_result[b] == -1) splitc_poll();
      }
    }
  }

  /*
   * The local value in the buckets array is the sum of all of
   * the previous buckets and the sum of the buckets of this
   * same value on previous processors
   */
  if (MYPROC == 0) {
    for (b = 0; b < RADIX; b++) {
      buckets[b] = sum + buckets[b];
      sum += scan_result[b];
    }
  }
  else {
    for (b = 0; b < RADIX; b++) {
      buckets[b] = sum + buckets[b] + scan_buffer[b];
      sum += scan_result[b];
    }
  }
}


void print_results(int np, int distrib)
{
  int i, j;

  on_one {
    printf ("Sort complete.\n\n");
    total_fill = total_scan = total_coallesce = 0;
    for (i = 0; i < DIGITS_PER_KEY; i++) { 
      for (j = 4; j > 0; j--)
        pass_time[i][j] -= pass_time[i][j - 1];
      printf ("Pass %d: ",i);
      printf ("Fill: %.4f s Histogram %.4f s Scan: %.3f s  Coallesce: %.3f s\n",
              pass_time[i][1], pass_time[i][2], pass_time[i][3],
              pass_time[i][4]);
      total_fill += pass_time[i][1];
      total_hist += pass_time[i][2];
      total_scan += pass_time[i][3];
      total_coallesce += pass_time[i][4];
    } 

    /* 
     * Print all on-one line  
     */
    i = 0;
    printf ("XXX Procs %d np: %d passes: %d dist: %d Fill: %.3f s  Histo: %.3f Scan: %.3f s  Coallesce: %.3f s Total %.3f\n",
            PROCS, np, DIGITS_PER_KEY,distrib,
            total_fill,total_hist,total_scan,total_coallesce,
            time_finish - time_start);
    
    printf ("\nTotal Time: %.3f\n\n", time_finish - time_start);
    printf ("MSorts per second: %.3f\n", (np*PROCS)/
            ((time_finish - time_start)*1000000));
    printf ("KSorts per second per processor: %.3f\n",
            np/((time_finish - time_start)*1000));
  }
}


void splitc_main(int argc, char *argv[])
{
  int pass, i, j;
  int np, distrib = 0;
  int procs, log_keys;
  unsigned (* spread g_keys)[NP_MAX];
  int c;
  extern char *optarg;

  np = NP_MAX;
  procs = PROCS;

  /* parse command line args */
  while ((c = getopt(argc, argv, "n:d:h")) != -1) {
    switch(c) {
    case 'h':
        on_one{
          printf("Radix Sort options: (defaults)\n");
          printf("  -n number of keys (in K) per processor  (512K)\n");
          printf("  -p number of procesors\n");
          printf("  -d key distribution\n");
        }
        exit(0);

      case 'n':  np = atoi(optarg);  np *= 1024; break;
      case 'd':  distrib = atoi(optarg); break;
    }
  }

  /* Calculate log_keys = log(np) */
  log_keys = 0;
  for (i = 1; i < np; i = i<<1)
    log_keys++;

  /* Print out configuration */
  on_one {
    printf ("\nSorting %d 32-bit keys on %d processors (%d keys/proc).\n", 
            np*PROCS, PROCS, np);
    printf ("Radix: %d (%d bits)\n\n", RADIX, BITS_PER_DIGIT);
  }
  

  /* Set up local pointers to global arrays. */
  keys = (unsigned *)g_keys0;
  buckets = (unsigned *)g_buckets;
  scan_buffer = (int *)g_scan_buffer;
  scan_result = (int *)g_scan_result;
  create_keys(np, keys, distrib);
  barrier();
  

  /* 
   * Sort the numbers, timing individual components.
   * A pass is made over each of the digits of the key.
   */
  time_start = get_seconds();
  for (pass = 0; pass < DIGITS_PER_KEY; pass++) {

    /*
     * Initialize the buckets
     */
    pass_time[pass][0] = get_seconds();
    keys = (unsigned *) (pass&1 ? tolocal(g_keys1) : tolocal(g_keys0));
    init_buckets();

    /* 
     * Count the number of keys with each value of a digit. 
     */
    pass_time[pass][1] = get_seconds();
    histogram(np, pass, keys);

    /* 
     * Use the bucket histogram to find global offsets. 
     */
    pass_time[pass][2] = get_seconds();
    all_scan_buckets();
    barrier();

    /* 
     * Move the keys to their new positions. 
     */
    pass_time[pass][3] =  get_seconds();
    g_keys = (pass&1) ? (unsigned *spread)g_keys0 : (unsigned *spread)g_keys1;
    all_coalesce(np, log_keys, pass, keys, g_keys); 

    pass_time[pass][4] = get_seconds(); 
    time_finish = get_seconds();
  }

  keys = (unsigned *)(pass & 1 ? tolocal(g_keys1) : tolocal(g_keys0));
  all_check_results(np, keys, 0);
  print_results(np, distrib);
}

