/* 
 * sample.sc - Split-C code for sample sort
 */
#include <stdio.h>
#include <split-c/split-c.h>
#include <split-c/com.h>
#include <split-c/atomic.h>


/*
 * Space for the samples from which the splitters are selected
 */
#define OVERSAMPLING    64
int g_sample1[PROCS]::[PROCSMAX*OVERSAMPLING];
int g_sample2[PROCS]::[PROCSMAX*OVERSAMPLING];

/* 
 * Splitter array -- identical on each processor
 */
int g_splitter[PROCS]::[PROCSMAX+1]; 
int *splitter;

/*
 * The destination processor of keys matching the splitters 
 */
int duplicate[PROCSMAX];

/*
 * The number of keys sent to this processor
 */
int my_num_keys;

/*
 * The number of keys worth of space allocated on each processor       
 */
#define EXPANSION(x)    (int)(1.6*x)
int bucket_size;

/*
 * The time of the entire algorithm and the individual phases
 */
double phase_time[6];
double time_finish, time_start;


/*
 * Procedure: print_results
 * Parameter:
 *   np (input): the number of keys per processor
 */
void print_results(int np)
{
  int max_keys, i;
  double expansion_factor;

  max_keys = all_reduce_to_one_max(my_num_keys);
  on_one {
    expansion_factor = (double)max_keys / (double)np;
    printf("XXX Procs: %d n/p: %d expansion %.3f ", PROCS, np,
           expansion_factor);
    for (i = 0; i < 3; i++)
      printf("Phase %d-%d: %.4lf ", i, i+1, phase_time[i+1]-phase_time[i]);
    printf("local_11bit_rsort %.4lf ",time_finish-phase_time[i]);
    printf("Total_Time: %.3f\n", time_finish - time_start);
  }
}

/*
 * Procedure: duplicate_splitters
 * Parameter:
 *   splitter (input): The list of P splitters
 *
 * duplicate[i] is set to the processor that this processor should
 * send a key that matches splitter[i]; every processor has a
 * different duplicate array so that if there are a lot of these
 * duplicate keys, then they are sent to different processors. 
 */
void duplicate_splitters()
{
  int p, i, a, first;
  
  /*
   * If there are not a sequence of identical splitters, then
   * if a key matches a duplicate, the even processors send the
   * key to the lower processor; the odd processors to the higher
   * processor.
   */
  for (p = 1; p < PROCS; p++) 
    duplicate[p] = p-1 + MYPROC%2;

  /*
   * Look for identical splitters.
   */
  a = splitter[1];
  for (p = 1; p < PROCS; ) {
    first = p;
    while (splitter[p] == a) p++;
    /*
     * There are (p-first+1) identical splitters.
     * Divide the processors into (p-first+1) groups and
     * have each group send to a different one of the (p-first+1) processors
     */
    for (i = first; i < p; i++)
      duplicate[i] = first-1 + MYPROC%(p - first + 1);
    a = splitter[p];
    p++;
  }
}


/*
 * Procedure: sel_bucket
 * Parameters:
 *   v (input): the key
 *   splitter (input): the list of splitters which determine
 *                     the destination processor of this key
 *
 * Returns: the destination processor of this key.
 */
static inline int sel_bucket(int v)
{
  int m, l=0, h=PROCS;
  int tmp;

  while(l+1 < h) {
    m = (l+h)>>1;
    if (v > (tmp = splitter[m])) l = m;
    else if (v < tmp) h = m;
    else return duplicate[m];
  }
      
  return l;
}

/*
 * Procedure: push_atomic1
 * Parameters:
 *   p (input): the destination buffer on this processor
 *   key1 (input): the key to save
 *
 * The destination buffer acts as a stack; my_num_keys is
 * the number of keys in the stack.  We have only allocated
 * space for bucket_size keys, so we must check that we don't overflow
 * our buffer space.
 */
void push_atomic1(int *p, int key1)
{
  p[my_num_keys] = key1;
  if (my_num_keys < bucket_size) my_num_keys++;
}


void sample_sort(unsigned int np, int *in, int *out)
{
  static int sample[OVERSAMPLING];
  int *all_samples1, *all_samples2, *global sample_ptr;
  int *nextKey,*addr, i, j;
  register int p;

  /* 
   * Every processor selects every np/OVERSAMPLE-th key
   * as its sample set; processor 0 puts these keys directly
   * in the all_samples buffer; the other processor put the
   * samples in a tempory buffer
   */
  phase_time[0] = get_seconds();
  if (MYPROC == 0) {
    all_samples1 = (int *)g_sample1;
    all_samples2 = (int *)g_sample2;
    for(i = 0; i < OVERSAMPLING; i++) 
      all_samples1[i] = in[i * (np / OVERSAMPLING)];
  }
  else {
    addr = (int *)g_sample1;
    addr += MYPROC*OVERSAMPLING;
    sample_ptr = toglobal(0, addr);

    for(i = 0; i < OVERSAMPLING; i++) 
      sample[i] = in[i * (np / OVERSAMPLING)];
  }
  barrier();

  /*
   * All processors now send the samples to processor 0 
   * The problem is, CMAM has a limit of 256 segments that can
   * be open at one time; therefore, we only let 256 processors
   * send at once.
   */
  for (i = 0; i < PROCS; i += 256) {
    if ((i < MYPROC)  && (MYPROC < (i+256))) 
      bulk_store(sample_ptr, sample, sizeof(sample));
    all_store_sync();
  }

  /* 
   * Sort samples on processor 0, select every OVERSAMPLING-th sample as
   * the splitters, and broadcast the all the splitters to all processors.
   */
  phase_time[1] = get_seconds();
  if (MYPROC == 0) {
    local_radix_sort_8bit(PROCS*OVERSAMPLING, all_samples1, all_samples2);
    splitter[0] = 0x80000000;   
    for(i = 1; i < PROCS; i++)
      splitter[i] = all_samples1[i*OVERSAMPLING];
  }
  my_num_keys=0;
  all_bcast_buffer(toglobal(MYPROC, splitter), sizeof(int)*PROCS);
  splitter[PROCS] = 0x7fffffff;
  /*
   * Determine the destination processor for those keys that match
   * one of the splitters.
   */
  duplicate_splitters();


  /*
   * We look at each of our keys, determine its destination processor,
   * and push the key onto the destination processor's local stack.
   */
  phase_time[2] = get_seconds();
  for(i = 0; i < np; i++) {
    p = sel_bucket(in[i]);
    atomic(push_atomic1, toglobal(p,out), in[i]); 
  }
  all_atomic_sync();

  /*
   * After we have received all of our keys, perform a local sort on
   * them; keys end up in "in" buffer.
   */
  phase_time[3] = get_seconds();
  local_radix_sort(my_num_keys, out, in);
}                             


void splitc_main(int argc, char**argv)
{
  int c;
  extern char *optarg;
  int i, dist=0, *in, *out, np = 512;

  /* Get options from the user */
  while ((c = getopt(argc, argv, "n:d:h")) != -1) {
    switch(c) {
    case 'h':
      on_one {
        printf("Sample sort options: (defaults)\n");
        printf("  -n number of keys (in K) per processor  (64K)\n");
        printf("  -d key distribution\n");
      }
      break;
    case 'n':  np = atoi(optarg);  np *= 1024; break;
    case 'd':  dist = atoi(optarg); break;
    }
  }
  on_one printf("Sorting %d ints (%d ints/proc)\n", PROCS*np, np);
  
  /* Allocate and initialize arrays */
  splitter = (int *)g_splitter;
  bucket_size = EXPANSION(np);
  in = (int *)all_spread_malloc(PROCS, sizeof(int) * bucket_size);
  out = (int *)all_spread_malloc(PROCS, sizeof(int) * bucket_size);
  if ((in == NULL) || (out == NULL)) {
    printf("proc %d: all_spread_malloc failed! exiting.\n",MYPROC);
    exit(-1);
  }
  
  create_keys(np, in, dist);
  barrier();
  time_start= get_seconds();
  sample_sort(np, in, out);
  barrier();
  time_finish=get_seconds();

  all_check_results(my_num_keys, in, 0);
  print_results(np);
}

