/*
 * transpose.sc - Routines to map between blocked and cyclic layouts of
 *   data across processors
 *
 */
#include <stdio.h>
#include <split-c/split-c.h>
#include "transpose.h"

/*
 * Procedure: gather
 * Parameters:
 *   np (input) keys per processor 
 *   s (input) list of n keys to be transposed
 *   d (output) keys gathered such that contiguous
 *              elements are sent to the same processor
 */
void gather(int np, int *d, int *s)
{
  int p, p2, j, i;
  int band = np / PROCS;

  /*
   * band number of keys are destined for each processor from this processor; 
   * stride through successive elements of s (i.e. get the first key destined
   * for each processor first, then the second key destined for each, etc.) 
   * and place in the scattered positions of d.
   */
  for (i = 0, j = 0; i < band; i++, j += PROCS)
    for (p = 0, p2 = 0; p < PROCS; p++, p2 += band)
      d[p2+i] = s[j + p];
}


/*
 * Procedure: all_transpose
 * Parameters: 
 *   np (input) keys per processor 
 *   s (input) the gathered list of np keys
 *   d (output) the keys we have been sent
 */
void all_transpose(int np, int *d, int *s)
{
  int *global dst;
  int p, p2, j, i, *index;
  int tr_offset, band;

  band = np / PROCS;
  tr_offset = MYPROC * band;
  index = d + tr_offset;

  barrier();
  /*
   * Pairs of processors exchange band number of keys on each iteration.
   */
  for (p2 = 0, p = MYPROC; p2 < PROCS; p2++, p = p2^MYPROC) {
    dst = toglobal(p, index);
    bulk_store(dst, s+p*band, sizeof(int) * band); 
    store_sync(sizeof(int)*band);  
  }
}


/*
 * Procedure: scatter
 * Parameters:
 *   np (input): number of keys per processor
 *   d (output): the keys in s locally transposed
 *   s  (input): the keys to be scattered
 */
void scatter(int np, int *d, int *s)
{
  int *global dst;
  int p, p2, i, *offset;
  int band;

  band = np / PROCS;
  offset = d + MYPROC*band;

  for (p = 0; p < PROCS; p++)
    for (p2 = p, i = 0; i < band; i++,  p2+=PROCS )
      d[p2] = *s++;
}
  
